#include <iostream>
#include <stdio.h>
#include <termio.h>
#include <unistd.h>
#include <dynamixel.h>
#include <math.h>
#include <stdio.h> // standard input / output functions
#include <string.h> // string function definitions
#include <unistd.h> // UNIX standard function definitions
#include <fcntl.h> // File control definitions
#include <stdlib.h>
#include <iostream>
#include <SerialStream.h>
#include <fcntl.h>
#include <cstdlib>
#include <sstream>
#include <stdexcept>

#include <ros/ros.h>
#include <geometry_msgs/PoseArray.h>
#include <geometry_msgs/Pose.h>
#include "std_msgs/String.h"

using namespace LibSerial;
using namespace std;

#define L						5

// Control table address
#define P_GOAL_POSITION_L		30
#define P_GOAL_POSITION_H		31
#define P_MOVING_SPEED_L		32
#define P_MOVING_SPEED_H		33
#define P_PRESENT_POSITION_L	36
#define P_PRESENT_POSITION_H	37
#define P_PRESENT_SPEED_L		38
#define P_PRESENT_SPEED_H		39
#define P_MOVING				46

//zero value
#define ZERO_SERVO1				2450
#define ZERO_SERVO2				1600
#define ZERO_SERVO4				440
#define ZERO_SERVO6				1800

class RoboticArm
{    
public:
	int *val_now; /**< */
	bool init; /**< true if the initialisation is done*/

	double l1; /**< Length of the first half of the robotic arm in cm*/
	double l2; /**< Length of the second half of the robotic arm in cm*/
	char *maplePort; /**< usb port connected to the Maple board*/
	double objectRadius; /**< radius of the object that will be moved*/

	bool isReady;
	
	ros::NodeHandle nh; /**< ROS node*/
	ros::Publisher chatter_pub; /**< ROS publisher that will be used to send debug informations*/
    ros::Subscriber sub; /**< ROS subscriber that will receive the command for the arm */
	
	/**
	 * Basic constructor
	 * Used for main tests
	 */ 
	RoboticArm(char *port)
	{
		ROS_INFO("helloooooo");
		l1 = 21.3;
		l2 = 20.4;
		maplePort =  port;//"ttyACM0";
		objectRadius = 3.5;
		isReady = false;
		
		sub = nh.subscribe("/arm_controler/arm_command", 10, &RoboticArm::armCommandCallback, this);
		chatter_pub = nh.advertise<std_msgs::String>("/robotic_arm/chatter", 1000);
		ROS_INFO("Arm initialised");
	}
	
	/**
	 * General constructor
	 * 
	 * @param l1 Length of the first half of the robotic arm in cm.
	 * @param l2 Length of the second half of the robotic arm in cm.
	 * @param maplePort usb port connected to the Maple board.
	 * @param objectRadius radius of the object that will be moved in cm.
	 */
	RoboticArm(double l1, double l2, char *maplePort, double objectRadius)
	{
		isReady = false;
		this->l1 = l1;
		this->l2 = l2;
		this->maplePort = maplePort;
		this->objectRadius = objectRadius;
		
		sub = nh.subscribe("/arm_controler/arm_command", 10, &RoboticArm::armCommandCallback, this);
		chatter_pub = nh.advertise<std_msgs::String>("/robotic_arm/chatter", 1000);
		ROS_INFO("Arm initialised");
	}
	
	~RoboticArm()
	{
	}
	
	/**
	* An object can be the object to move or an obstacle
	*/
	struct OBJECT {
		double x; /**< x position for the center of the object */
		double y; /**< y position for the center of the object */
		double z; /**< z position for the center of the object */
		double height; /**< height of the object */
		double radius; /**< radius or width of the object */
	};

	/*----------Maple mini Communication----------*/
	
	/**
	 * try to open the choosen port
	 * 
	 * @param portName name of a specific port
	 * @return the file descripiton of the specific port
	 */
	int open_port(char *portName)
	{
		int fd; /**<  file description for the serial port */
		const char *str = "/dev/";
		char path[20];
		strcpy(path,str);
		strcat(path, portName);
		fd = open(path, O_RDWR | O_NOCTTY | O_NDELAY);

		if(fd == -1) // if open is unsucessful
		{
			ROS_ERROR("open_port: Unable to open "/* << path*/);
			//throw exception("Port not found");
		}
		else
		{
			fcntl(fd, F_SETFL, 0);
		}

		return(fd);
	}

	/**
	 * Configure a port (baudrates, no parity, stop bits, data bits)
	 * 
	 * @param fd file description of a specific port
	 * @return the file descripiton of the specific port after being configurated
	 */
	int configure_port(int fd)
	{
		struct termios port_settings;     /**<  structure to store the port settings in*/

		cfsetispeed(&port_settings, B115200);    // set baud rates
		cfsetospeed(&port_settings, B115200);

		port_settings.c_cflag &= ~PARENB;    // set no parity, stop bits, data bits
		port_settings.c_cflag &= ~CSTOPB;
		port_settings.c_cflag &= ~CSIZE;
		port_settings.c_cflag |= CS8;

		tcsetattr(fd, TCSANOW, &port_settings);    // apply the settings to the port
		return(fd);

	}
	
	/**
	 * Send an input to the maple_mini.
	 * 
	 * @param fd file description of a specific port
	 * @param send_bytes bytes to send
	 */
	void query2maple(int fd, unsigned char send_bytes[])
	{
		write(fd, send_bytes, strlen((char*)send_bytes));

	}

	/**
	 * Send a message to the maple board in order to switch the hand state (open or close)
	 * 
	 * @param portNamefile usb port connected to the maple board
	 * @return state of the hand
	 */
	char change_hand_state(char *portName)
	{
		SerialStream serial_port ;
		serial_port.Open(portName) ;
		if ( ! serial_port.good() )
		{
			/*std::cerr << "[" << __FILE__ << ":" << __LINE__ << "] "
					<< "Error: Could not open serial port." << portName
					<< std::endl ;*/
			exit(1) ;
		}
		//
		// Set the baud rate of the serial port.
		//
		serial_port.SetBaudRate( SerialStreamBuf::BAUD_9600 ) ;
		if ( ! serial_port.good() )
		{
			/*std::cerr << "Error: Could not set the baud rate." <<  std::endl ;*/
			exit(1) ;
		}
		char output;
		char out_buf[] = "1";
		serial_port.write(out_buf, 1);
		while(1)
		{
			serial_port.get(output);
			break;
		}
		serial_port.Close();
		return output;
	}
	
	/**
	 * Send a message to the maple board in order to open the hand
	 * 
	 * @param portNamefile usb port connected to the maple board
	 */	
	void open_hand(char *portName)
	{
		const char *str = "/dev/";
		char path[20];
		strcpy(path,str);
		strcat(path, portName);

		char output;
		output = change_hand_state(path);
		if (output != '0')
		{
			output = change_hand_state(path);
		}
	}

	/**
	 * Send a message to the maple board in order to close the hand
	 * 
	 * @param portNamefile usb port connected to the maple board
	 */
	void close_hand(char *portName)
	{
		const char *str = "/dev/";
		char path[20];
		strcpy(path,str);
		strcat(path, portName);

		char output;
		output = change_hand_state(path);
		if (output != '1')
		{
			output = change_hand_state(path);
		}
	}

	/*----------Dynamixel Communication----------*/
	
	/**
	 * Convert the required angles to order for the servomotors
	 * 
	 * @param angle array that contain the angle in degrees for each servomotors. 
	 * angle[0] is the alpha angle for the servo of id 1 (EX106)
	 * angle[1] is the beta angle for the servo of id 4 (RX64)
	 * angle[2] is the theta angle for the servo of id 2 (EX106)
	 *
	 * @return angle value to send to the servomotors
	 */
	int * convert_servo_val(double * angle)
	{
		static int val[3];
		double temp;

		temp = angle[0]*836.3632+ZERO_SERVO1;
		if (angle[0]<0)
			temp += 100*angle[0];
		else
			temp += 50*angle[0];
		val[0] = (int)temp;

		temp = -(3.1416-angle[1])*195.3786+ZERO_SERVO4 +10; //calibrage
		val[1] = (int)temp;

		temp = -angle[2]*836.3632+ZERO_SERVO2 +10;//calibrage
		val[2] = (int)temp;

		return val;
	}

	/**
	 * Calculate the angle to send to the servomotor 6 (wrist) in order to have an horizontal position for the hand
	 * 
	 * @param servo_2 angle of the servomotor 2 (servomotors value)
	 * @param servo_4 angle of the servomotor 4 (servomotors value)
	 *
	 * @return angle value to send to the servomotor 6
	 */	
	int servo_6_cal(int servo_2, int servo_4)
	{
		double servo_6;
		double servo_4_angle = (servo_4 - ZERO_SERVO4)/195.3786 + 3.1416;
		double servo_2_angle = (ZERO_SERVO2 - servo_2)/836.3632;
		servo_6 = 651.8971*(3.1416 - servo_2_angle - servo_4_angle) + ZERO_SERVO6;

		return (int)servo_6+30+300;
	}

	/**
	 * Move the hand to a specific position
	 *
	 *@param val final angles to send to the servomotors
	 *@param type type of movement to do
	 *type 1 : initialization, the arm will move near to the objective and then go to the objective
	 *type 2 : move an object to the target position, same as previously
	 *type 3 : take an object the arm will move to a position at the right of the wanting position and will move to the left. Like this, we will avoid colliding with the object;
	 */
	void movement(int *val, int type)
	{
		//Pour éviter la collision avec la table, le servo 4 la ouvrir en 2 phases: 1ere phase: ouvrir à le point qui est = bon point - 300(servo 2 en plus haut)
		// et puis le servo 2 va descendre en meme temp avec le servo 4.

		//calculer la postition ciblee du servo 6
		double pos_6 = servo_6_cal(1000,0);
		//Fermer et préparer le bras
		//ROS_INFO("Movement Type = " << type);
		//ROS_INFO("pos_6 init = " << pos_6);
		switch(type)
		{
			case 1:
				pos_6 = 3600;
				dxl_write_word( 6, P_MOVING_SPEED_L,400);
				dxl_write_word( 6, P_GOAL_POSITION_L,pos_6);

				dxl_write_word( 2, P_GOAL_POSITION_L,1150);
				dxl_write_word( 4, P_GOAL_POSITION_L,0);
				dxl_write_word( 6, P_MOVING_SPEED_L,50);
				break;
			case 2:
			case 3:
				dxl_write_word( 2, P_GOAL_POSITION_L,1150);
				dxl_write_word( 4, P_GOAL_POSITION_L,0);
				dxl_write_word( 6, P_GOAL_POSITION_L,pos_6);
				usleep(500000);
				break;
		}
		//bouger le 1er servo
		if(type == 3)
		{
			dxl_write_word( 1, P_GOAL_POSITION_L,val[0] + 200);
			//ROS_INFO("move to the right of the object");
		}
		else
		{
			dxl_write_word( 1, P_GOAL_POSITION_L,val[0]);
			//ROS_INFO("move to the object");
		}

		//phase 1 : le servo 4 ouvre au bon endroit - 75, servo 5 tourner a droit un peu pour eviter de toucher l'objet
		pos_6 = servo_6_cal(1000,val[1]-150)+50;
		//ROS_INFO("pos_6 1st movement = " << pos_6);
		dxl_write_word( 6, P_MOVING_SPEED_L,50);
		dxl_write_word( 6, P_GOAL_POSITION_L,pos_6);
		if (val[1]-75 <0)
			dxl_write_word( 4, P_GOAL_POSITION_L,0);
		else
			dxl_write_word( 4, P_GOAL_POSITION_L,val[1]-75);
		if (type ==1)
			dxl_write_word( 5, P_GOAL_POSITION_L,600);
		usleep(2000000);
		//finir phase 1

		//phase 2 : go to the right place
		//pos_6 = servo_6_cal(val[2],val[1]);
		//ROS_INFO("pos_6 2nd movement = " << pos_6);
		dxl_write_word( 6, P_MOVING_SPEED_L,400);
		dxl_write_word( 6, P_GOAL_POSITION_L,pos_6);
		dxl_write_word( 2, P_MOVING_SPEED_L,100);
		//dxl_write_word( 2, P_GOAL_POSITION_L,val[0]+200);
		//usleep(200000);
		//dxl_write_word( 2, P_GOAL_POSITION_L,val[0]+100);
		//usleep(200000);
		dxl_write_word( 2, P_GOAL_POSITION_L,val[2]);
		dxl_write_word( 4, P_MOVING_SPEED_L,50);
		dxl_write_word( 4, P_GOAL_POSITION_L,val[1]-30);
		if (type ==1)
		{
			dxl_write_word( 5, P_MOVING_SPEED_L,30);
			usleep(200000);
			dxl_write_word( 5, P_GOAL_POSITION_L,580);
			usleep(200000);
			dxl_write_word( 5, P_GOAL_POSITION_L,560);
			usleep(200000);
			dxl_write_word( 5, P_GOAL_POSITION_L,550);
		}
		dxl_write_word( 4, P_MOVING_SPEED_L,30);
		usleep(200000);
		dxl_write_word( 4, P_GOAL_POSITION_L,val[1]-20);
		usleep(200000);
		dxl_write_word( 4, P_GOAL_POSITION_L,val[1]-10);
		usleep(200000);
		dxl_write_word( 4, P_GOAL_POSITION_L,val[1]);
		dxl_write_word( 2, P_MOVING_SPEED_L,50);
		dxl_write_word( 4, P_MOVING_SPEED_L,100);
		usleep(500000);

		if(type == 3)
		{
			dxl_write_word( 1, P_MOVING_SPEED_L,30);
			usleep(1000000);
			dxl_write_word( 1, P_GOAL_POSITION_L,val[0]-20);
			usleep(1000000);
			dxl_write_word( 1, P_MOVING_SPEED_L,100);
		}
		val_now = val;
	}

	
	/**
	 * Move the arm to the intial position
	 *
	 */
	void initialization()
	{
		dxl_write_word( 1, P_MOVING_SPEED_L,200);
		dxl_write_word( 2, P_MOVING_SPEED_L,150);
		dxl_write_word( 4, P_MOVING_SPEED_L,200);
		dxl_write_word( 6, P_MOVING_SPEED_L,400);

		dxl_write_word( 2, P_GOAL_POSITION_L,1150);
		ROS_INFO("Servo 2 moved to intial position");	
		usleep(1000000);
		
		dxl_write_word( 6, P_GOAL_POSITION_L,3100);
		ROS_INFO("Servo 6 moved to intial position");
		dxl_write_word( 4, P_GOAL_POSITION_L,0);
		ROS_INFO("Servo 4 moved to intial position");
		dxl_write_word( 1, P_GOAL_POSITION_L,4000);//dxl_write_word( 1, P_GOAL_POSITION_L,1000);
		ROS_INFO("Servo 1 moved to intial position");

		//changer la vitesse
		dxl_write_word( 1, P_MOVING_SPEED_L,100);
		dxl_write_word( 2, P_MOVING_SPEED_L,100);
		dxl_write_word( 4, P_MOVING_SPEED_L,100);
		dxl_write_word( 6, P_MOVING_SPEED_L,50);
		init = true;

		static int val[3];
		val[0] = 4000;//val[0] = 1000;
		val[1] = 0;
		val[2] = 1150;
		val_now = (int *)val;

		open_hand(maplePort);
		isReady = true;
	}
	
	/**
	 * You shall not talk about this secret function
	 */
	void secretFunction()
	{
		dxl_write_word( 1, P_MOVING_SPEED_L,300);
		dxl_write_word( 2, P_MOVING_SPEED_L,300);
		dxl_write_word( 4, P_MOVING_SPEED_L,300);
		dxl_write_word( 6, P_MOVING_SPEED_L,500);

		dxl_write_word( 2, P_GOAL_POSITION_L,1150);
		dxl_write_word( 4, P_GOAL_POSITION_L,500);
		dxl_write_word( 6, P_GOAL_POSITION_L,2500);
	}
	/*---------- Distance operations ----------*/

	/**Calculate the position of the arm in order to access to the object on a specific position
	 * 
	 * @param x x coordinate of the object
	 * @param y y coordinate of the object
	 * @param h height of the object from the top of the arm (top of servomotors 1)
	 * @param size radius or width of the object
	 * 
	 * @return angles values (in rad) for the 3 servomotors used for moving the arm
	 */
	double * IK(double x, double y, double h, double size)
	{
		static double angle[3];
		double alpha;
		double longeur_bis;
		double l_prim;
		double longeur;
		double a2;
		//modification of the coordinates
		longeur = sqrt(x*x + y*y);
		alpha = asin(size/longeur);// alpha correspond a l'ecart de l'angle tel que la main sera sur le cote de l'objet
		longeur_bis = sqrt(longeur*longeur-size*size)-L-2.5; // Nouvelle longueur tel que le centre de la main soit sur le cote de l'objet
		if (y == 0)
			if (x > 0)
				angle[0] = 1.5708;
			else
				angle[0] = -1.5708;
		else
			angle[0] = atan(x/y);
		angle[0] += alpha;

		//calculer les angles
		l_prim = sqrt(h*h + longeur_bis*longeur_bis);
		if (longeur == 0)
			a2 = 1.5708;
		else
			a2 = atan(h/longeur_bis);
		angle[2] = acos((l1*l1 + l_prim*l_prim-l2*l2)/(2*l1*l_prim)) - a2;
		angle[1] = acos((l1*l1 + l2*l2 - l_prim*l_prim)/(2*l1*l2));
		return angle;
	}

	/**Check if the arm can reach the object
	 * 
	 * @param x x coordinate of the object
	 * @param y y coordinate of the object
	 * @param h height of the object from the top of the arm (top of servomotors 1)
	 * @param radius radius or width of the object
	 * 
	 * @return true if the object is reachable, else false
	 */
	bool check_values(double x, double y, double h, double radius)
	{
		//y ne peut pas etre negative
		if (y<8.8)
		{
			ROS_INFO("y value must be more than 8.8cm/n");
			return false;
		}

		double longeur = sqrt(x*x+y*y);
		double longeur_bis = sqrt(longeur*longeur-radius*radius)-L-2.5;

		if (longeur_bis > sqrt ((l1+l2)*(l1+l2) - h*h))
		{
			ROS_INFO("The position can't be reach because values are too big\n");
			return false;
		}

		double* angle_init = IK(x, y, h, radius);
		if ((angle_init[2] > (ZERO_SERVO2-1000)/836.3632) || (angle_init[1] < -ZERO_SERVO4/195.3786+3.1416))
		{
			ROS_INFO("The position can't be reach because values are too short\n");
			return false;
		}

		return true;
	}

	/**Check if the arm can reach whithout knocking in obstacles
	 * 
	 * @param x x coordinate of the object
	 * @param y y coordinate of the object
	 * @param h height of the object from the top of the arm (top of servomotors 1)
	 * @param radius radius or width of the object
	 * @param obstacles list of the detected obstacles
	 * 
	 * @return true if the object is reachable, else false
	 */
	bool check_obstacles(double x, double y, double h, double radius, OBJECT obstacles[])
	{
		return true;
	}

	// TO DO : move an object whithout hurting an obstacle or the ceilling
	void simple_move_object(OBJECT target, OBJECT obstacles[], double ceiling_height, double new_x, double new_y)
	{
		//TO DO
	}

	/**Move an object from a coordinate to another coordinate
	 * 
	 * @param old_x x coordinate of the object
	 * @param old_y y coordinate of the object
	 * @param old_z height of the object from the top of the arm (top of servomotors 1)
	 * @param new_x x coordinate of the target position
	 * @param new_y y coordinate of the target position
	 * @param new_z height of the target position from the top of the arm (top of servomotors 1)
	 */
	bool move_object(int old_x, int old_y, int old_z, int new_x, int new_y, int new_z)
	{ 
		int *val;
		double *angle;
	
		if(!check_values(old_x, old_y, old_z, objectRadius) || !check_values(new_x, new_y, new_z, objectRadius))
		{
			return false;
		}
		
		isReady = false;
		angle = IK(old_x, old_y, old_z, objectRadius);
		val = convert_servo_val(angle);
		//We move the arm to the selected position
		movement(val,3);
		sleep(2);
		close_hand(maplePort);
		ROS_INFO("Object taken in (%ld,%ld)",old_x, old_y);
		sleep(2);
		angle = IK(new_x, new_y, new_z, objectRadius);
		val = convert_servo_val(angle);
		movement(val,2);
		sleep(2);
		open_hand(maplePort);
		ROS_INFO("Object moved to (%ld,%ld)", new_x, new_y);
		usleep(1000000);
		dxl_write_word( 1, P_MOVING_SPEED_L,30);
		usleep(200000);
		dxl_write_word( 1, P_GOAL_POSITION_L,val[0]+200);
		usleep(1000000);
		dxl_write_word( 1, P_MOVING_SPEED_L,100);
		return true;
	}

	/**
	 * Callback for the subscriber sub. It will process the command from /arm_controler/arm_command
	 * 
	 * @param msg command for the arm from the arm controler
	 */
	
	void armCommandCallback(const geometry_msgs::PoseArray msg)
	{
		double xCoeff = 90;
		double yCoeff = 75;
		double zCoeff = 100;
		
		if(isReady)
		{
			std_msgs::String msgPub;
			std::stringstream ss;
			if(move_object(msg.poses[0].position.x * xCoeff, msg.poses[0].position.y * yCoeff, /*msg.poses[0].position.z*zCoeff*/30, msg.poses[1].position.x * xCoeff, msg.poses[1].position.y*yCoeff,/* msg.poses[1].position.z*zCoeff*/30))
			{
				ss << "x_object : " << floor(msg.poses[0].position.x * 100.)/100. << ", y_object : " << floor(msg.poses[0].position.y *100. )/100. << ", z_object : " << floor(msg.poses[0].position.z*100.)/100.;
				ss << "\n -> x_target : " << floor(msg.poses[1].position.x * 100.)/100. << ", y_target : " << floor(msg.poses[1].position.y *100. )/100. << ", z_target : " << floor(msg.poses[1].position.z*100.)/100.;
				initialization();
			}
			else
			{
				ss << "object or target can't be reached";
			}
			msgPub.data = ss.str();
			chatter_pub.publish(msgPub);
		}
		else
		{
			ROS_INFO("ARM IS NOT READY");
		}
			
	}
};

int main( int argc, char **argv)
{
    //initial values for test
    double height = 30;

    double x_object;
    double y_object;
    double x_target;
    double y_target;

    int baudnum = 1;
    int deviceIndex = 0;

	ros::init(argc, argv, "robotic_arm");
	
	RoboticArm ra("ttyACM0");
	
    ROS_INFO("================ Arm control for Linux ===================" );
	
    ///////// Open USB2Dynamixel ////////////
    if( dxl_initialize(deviceIndex, baudnum) == 0 )
    {
        ROS_INFO("Failed to open USB2Dynamixel!\n" );
        return 0;
    }
    else
    {
        ROS_INFO("Succeed to open USB2Dynamixel!\n" );
    }

    ROS_INFO("==================== Initialization ======================");

    ra.initialization();
    ROS_INFO("Initialisation done");
    sleep(2);
    
    ROS_INFO("======================= Movement =========================");

	/*ros::Rate r(10);
    while(ros::ok())
    {
		ros::spinOnce();
		r.sleep();
    }*/
    
	ros::spin();
	
    return 0;
}
